from actionpack import Action


class FakeAction(Action):

    result = f'Performing Action.'

    def __init__(self, name=None, exception=None):
        self._name(name)
        self.exception = exception
        self.state = {'this': 'state'}

    def instruction(self):
        return self.result

    def validate(self):
        if self.exception:
            raise self.exception
        return super().validate()




from oslash import Left
from oslash import Right
from typing import List
from typing import Union


class Action:
    def perform(self) -> Union[Left, Right]:
        if callable(self.instruction):
            try:
                return Right(self.validate().instruction())
            except Exception as e:
                return Left(e)
        else:
            return Left(TypeError(f'Must be callable: {self.instruction}'))

    def validate(self):
        return self

    def _name(self, _name):
        self.name = _name

    def __getstate__(self):
        return vars(self)

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            raise self.__class__.NotComparable(f'{str(self)} ≠ {str(other)}')

        return self.__getstate__() == other.__getstate__()

    def __repr__(self):
        return f'{self.__class__.__name__}'

    class NotComparable(Exception): pass




from actionpack import Action
from pathlib import Path


class ReadBytes(Action):
    def __init__(self, filename: str):
        self.path = Path(filename)

    def instruction(self) -> bytes:
        return self.path.read_bytes()

    def validate(self):
        if not self.path.exists():
            raise FileNotFoundError(str(self.path))
        if self.path.is_dir():
            raise IsADirectoryError(str(self.path))
        return self




import pickle

from actionpack.actions import ReadBytes
from actionpack.utils import pickleable

from oslash import Left
from oslash import Right
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch


class ReadBytesTest(TestCase):

    contents = 'some file contents.'.encode()

    def setUp(self):
        self.action = ReadBytes(__file__)

    @patch('pathlib.Path.read_bytes')
    def test_can_ReadBytes(self, mock_input):
        mock_input.return_value = self.contents
        result = self.action.perform()
        self.assertIsInstance(result, Right)
        self.assertEqual(result.value, self.contents)

    @patch('pathlib.Path.read_bytes')
    def test_can_ReadBytes_even_if_failure(self, mock_input):
        mock_input.return_value = self.contents

        invalid_file_result = ReadBytes('some/invalid/filepath').perform()
        self.assertIsInstance(invalid_file_result, Left)
        self.assertIsInstance(invalid_file_result.value, FileNotFoundError)

        directory_result = ReadBytes(Path(__file__).parent).perform()
        self.assertIsInstance(directory_result, Left)
        self.assertIsInstance(directory_result.value, IsADirectoryError)

    def test_can_pickle(self):
        pickled = pickleable(self.action)
        unpickled = pickle.loads(pickled)

        self.assertTrue(pickleable(self.action))
        self.assertEqual(unpickled.__dict__, self.action.__dict__)




import pickle

from actionpack.actions import MakeRequest
from actionpack.utils import pickleable
from tests.actionpack.actions import FakeResponse

from requests import Response
from requests.exceptions import MissingSchema
from oslash import Left
from oslash import Right
from unittest import TestCase
from unittest.mock import patch
from validators import ValidationFailure


class MakeRequestTest(TestCase):

    def setUp(self):
        self.action = MakeRequest('GET', 'http://localhost')

    @patch('requests.Session.send')
    def test_can_MakeRequest(self, mock_session_send):
        mock_session_send.return_value = FakeResponse()
        self.assertIsInstance(self.action.perform(), Right)

    @patch('requests.Session.send')
    def test_can_MakeRequest_even_if_failure(self, mock_session_send):
        mock_session_send.return_value = FakeResponse(status=500)

        first_action = MakeRequest('SUP', 'http://localhost')
        first_result = first_action.perform()
        self.assertIsInstance(first_result, Left)
        self.assertIsInstance(first_result.value, ValueError)

        second_action = MakeRequest('GET', 'http://some/invalid/url')
        second_result = second_action.perform()
        self.assertIsInstance(second_result, Left)
        self.assertIsInstance(second_result.value, ValidationFailure)

    def test_can_pickle(self):
        pickled = pickleable(self.action)
        unpickled = pickle.loads(pickled)

        self.assertTrue(pickleable(self.action))
        self.assertEqual(unpickled.__dict__, self.action.__dict__)




import pickle

from actionpack.actions import WriteBytes
from actionpack.utils import pickleable

from io import BytesIO
from io import TextIOWrapper
from unittest import TestCase
from unittest.mock import patch


class WriteBytesTest(TestCase):

    class FakeFile:
        def __init__(self, contents: bytes=bytes(), mode: str=None):
            self.buffer = BytesIO(contents)
            self.buffer.read()
            self.mode = mode

        def read(self):
            self.buffer.seek(0)
            return self.buffer.read()

        def write(self, data: bytes):
            if self.mode == 'wb':
                self.buffer.seek(0)
            self.buffer.write(data)
            return len(data)

        def __enter__(self):
            return self

        def __exit__(self, *args):
            self.closed = True

    def setUp(self):
        self.salutation = 'Hello.'.encode()
        self.question = ' How are you?'.encode()
        self.action = WriteBytes('valid/path/to/file', self.question)

    @patch('pathlib.Path.open')
    def test_can_WriteBytes(self, mock_output):
        file = self.FakeFile(self.salutation)
        mock_output.return_value = file
        result = self.action.perform()

        self.assertEqual(file.read(), self.salutation + self.question)
        self.assertEqual(result.value, len(self.question))

    @patch('pathlib.Path.open')
    def test_can_overWriteBytes(self, mock_output):
        file = self.FakeFile(self.salutation, 'wb')
        mock_output.return_value = file
        action = WriteBytes('valid/path/to/file', self.question, overwrite=True)
        result = self.action.perform()

        self.assertEqual(file.read(), self.question)
        self.assertEqual(result.value, len(self.question))

    def test_can_pickle(self):
        pickled = pickleable(self.action)
        unpickled = pickle.loads(pickled)

        self.assertTrue(pickleable(self.action))
        self.assertEqual(unpickled.__dict__, self.action.__dict__)




import pickle

from actionpack import Action
from actionpack.utils import pickleable
from tests.actionpack import FakeAction

from oslash import Left
from oslash import Right
from unittest import TestCase


class ActionTest(TestCase):

    def test_Action_produces_Right_result_when_performed(self):
        result = FakeAction().perform()
        self.assertIsInstance(result, Right)
        self.assertEqual(result.value, FakeAction.result)

    def test_Action_produces_Left_if_exception_raised_when_performed(self):
        exception = Exception('something went wrong :/')
        result = FakeAction(exception=exception).perform()
        self.assertIsInstance(result, Left)
        self.assertEqual(result.value, exception)

    def test_Action_can_be_serialized(self):
        action = FakeAction()
        pickled = pickle.dumps(action)
        unpickled = pickle.loads(pickled)

        self.assertEqual(pickleable(action), pickled)
        self.assertEqual(unpickled.result, action.result)
        self.assertEqual(unpickled.state, action.state)




import pickle

from itertools import chain
from typing import Optional


def tally(num=1):
    while (num if num > 0 else -1 * num) > 0:
        yield 1 if num > 0 else -1
        num = num - 1 if num > 0 else num + 1


def pickleable(obj) -> Optional[bytes]:
    try:
        return pickle.dumps(obj)
    except Exception:
        pass




from actionpack import Action

from requests import Request
from requests import Session
from validators import url as is_url



class MakeRequest(Action):

    methods = ('CONNECT', 'GET', 'DELETE', 'HEAD', 'OPTIONS', 'PATCH', 'POST', 'PUT', 'TRACE')

    def __init__(self, method: str, url: str, data: dict=None, headers: dict=None, session: Session=None):
        self.method = method.upper()
        self.url = url
        self.method = method
        self.data = data
        self.headers = headers
        self.session = session

    def prepare(self, method: str, url: str, data: dict=None, headers=None) -> Request:
        return Request(method, self.url, data=data, headers=headers).prepare()

    def instruction(self):
        request = self.prepare(self.method, self.url, self.data, self.headers)
        return (self.session if self.session else Session()).send(request)

    def validate(self):
        if self.method not in self.methods:
            raise ValueError(f'Invalid HTTP method: {self.method}')
        url_validation = is_url(self.url)
        if url_validation is not True:
            raise url_validation
        return self




from actionpack.utils import tally
from actionpack.utils import pickleable

from unittest import TestCase


class UtilsTest(TestCase):

    def test_tally(self):
        self.assertEqual(list(tally(0)), [])
        self.assertEqual(list(tally()), [1])
        self.assertEqual(list(tally(2)), [1, 1])

    def test_pickleable(self):
        self.assertTrue(pickleable(CanPickleMe()))
        self.assertFalse(pickleable(CannotPickleMe()))


class CanPickleMe: pass


class CannotPickleMe:
    def __init__(self):
        self.get_pickle_juice = lambda: 'here ya go'

